package com.chelvc.framework.base.context;

import java.util.Collection;
import java.util.Objects;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.function.Consumer;
import java.util.function.Supplier;

import com.chelvc.framework.base.model.Session;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.core.task.TaskExecutor;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

/**
 * Executor上下文工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
public class ExecutorContextHolder implements ApplicationContextAware {
    /**
     * 空任务结果数组
     */
    private static final CompletableFuture<?>[] EMPTY_FUTURES = new CompletableFuture<?>[0];

    /**
     * 多线程任务执行实例
     */
    private static Executor EXECUTOR;

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        EXECUTOR = applicationContext.getBean(TaskExecutor.class);
    }

    /**
     * 获取应用上下文中配置的Executor
     *
     * @return Executor对象
     */
    public static Executor getExecutor() {
        return Objects.requireNonNull(EXECUTOR, "Executor has not been initialized");
    }

    /**
     * 异步调用方法
     *
     * @param runnable Runnable实例
     * @return CompletableFuture实例
     */
    public static CompletableFuture<Void> run(@NonNull Runnable runnable) {
        return run(EXECUTOR, runnable);
    }

    /**
     * 异步调用方法
     *
     * @param executor 线程执行器
     * @param runnable Runnable实例
     * @return CompletableFuture实例
     */
    public static CompletableFuture<Void> run(Executor executor, @NonNull Runnable runnable) {
        long id = Thread.currentThread().getId();
        Session session = SessionContextHolder.getSession(true);
        Runnable function = () -> {
            if (id == Thread.currentThread().getId()) {
                runnable.run();
            } else {
                // 如果子、父线程不是同一个线程则设置会话上下文信息
                SessionContextHolder.setSession(session);
                try {
                    runnable.run();
                } finally {
                    SessionContextHolder.clearSessionContext();
                }
            }
        };
        CompletableFuture<Void> future = executor == null ? CompletableFuture.runAsync(function) :
                CompletableFuture.runAsync(function, executor);
        return future.exceptionally(e -> {
            log.error(e.getMessage(), e);
            return null;
        });
    }

    /**
     * 异步调用方法
     *
     * @param supplier 回调方法实例
     * @param <T>      任务结果类型
     * @return CompletableFuture实例
     */
    public static <T> CompletableFuture<T> supply(@NonNull Supplier<T> supplier) {
        return supply(EXECUTOR, supplier);
    }

    /**
     * 异步调用方法
     *
     * @param executor 线程执行器
     * @param supplier 回调方法实例
     * @param <T>      任务结果类型
     * @return CompletableFuture实例
     */
    public static <T> CompletableFuture<T> supply(Executor executor, @NonNull Supplier<T> supplier) {
        long id = Thread.currentThread().getId();
        Session session = SessionContextHolder.getSession(true);
        Supplier<T> function = () -> {
            if (id == Thread.currentThread().getId()) {
                return supplier.get();
            }

            // 如果子、父线程不是同一个线程则设置会话上下文信息
            SessionContextHolder.setSession(session);
            try {
                return supplier.get();
            } finally {
                SessionContextHolder.clearSessionContext();
            }
        };
        CompletableFuture<T> future = executor == null ? CompletableFuture.supplyAsync(function) :
                CompletableFuture.supplyAsync(function, executor);
        return future.exceptionally(e -> {
            log.error(e.getMessage(), e);
            return null;
        });
    }

    /**
     * 异步执行方法（等待执行完成）
     *
     * @param array Runnable实例数组
     */
    public static void execute(@NonNull Runnable... array) {
        execute(EXECUTOR, array);
    }

    /**
     * 异步执行方法（等待执行完成）
     *
     * @param executor 线程执行器
     * @param array    Runnable实例数组
     */
    public static void execute(Executor executor, @NonNull Runnable... array) {
        if (array != null && array.length == 1) {
            run(executor, array[0]).join();
        } else if (array != null) {
            CompletableFuture<?>[] futures = new CompletableFuture<?>[array.length];
            for (int i = 0; i < array.length; i++) {
                futures[i] = run(executor, array[i]);
            }
            CompletableFuture.allOf(futures).join();
        }
    }

    /**
     * 异步执行方法（等待执行完成）
     *
     * @param array 回调方法数组
     * @param <T>   任务结果类型
     * @return CompletableFuture实例
     */
    @SuppressWarnings("unchecked")
    public static <T> CompletableFuture<T>[] execute(@NonNull Supplier<T>... array) {
        return execute(EXECUTOR, array);
    }

    /**
     * 异步执行方法（等待执行完成）
     *
     * @param executor 线程执行器
     * @param array    回调方法数组
     * @param <T>      任务结果类型
     * @return CompletableFuture实例
     */
    @SuppressWarnings({"unchecked", "rawtypes"})
    public static <T> CompletableFuture<T>[] execute(Executor executor, @NonNull Supplier<T>... array) {
        if (array == null || array.length == 0) {
            return new CompletableFuture[0];
        }
        CompletableFuture[] futures = new CompletableFuture[array.length];
        for (int i = 0; i < array.length; i++) {
            futures[i] = supply(executor, array[i]);
        }
        CompletableFuture.allOf(futures).join();
        return futures;
    }

    /**
     * 等待任务执行完成
     *
     * @param futures 任务数组
     * @param <T>     任务结果类型
     */
    @SafeVarargs
    public static <T> void join(@NonNull CompletableFuture<T>... futures) {
        if (futures != null && futures.length > 0) {
            CompletableFuture.allOf(futures).join();
        }
    }

    /**
     * 等待任务执行完成
     *
     * @param futures 任务集合
     * @param <T>     任务结果类型
     */
    public static <T> void join(@NonNull Collection<CompletableFuture<T>> futures) {
        if (!CollectionUtils.isEmpty(futures)) {
            CompletableFuture.allOf(futures.toArray(EMPTY_FUTURES)).join();
        }
    }

    /**
     * 获取任务执行结果
     *
     * @param futures  任务数组
     * @param consumer 任务结果处理方法
     * @param <T>      任务结果对象类型
     */
    public static <T> void results(@NonNull CompletableFuture<T>[] futures, @NonNull Consumer<T> consumer) {
        if (futures != null && futures.length > 0) {
            join(futures);
            for (CompletableFuture<T> future : futures) {
                try {
                    consumer.accept(future.get());
                } catch (InterruptedException | ExecutionException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    /**
     * 获取任务执行结果
     *
     * @param futures  任务集合
     * @param consumer 任务结果处理方法
     * @param <T>      任务结果对象类型
     */
    public static <T> void results(@NonNull Collection<CompletableFuture<T>> futures, @NonNull Consumer<T> consumer) {
        if (!CollectionUtils.isEmpty(futures)) {
            join(futures);
            futures.forEach(future -> {
                try {
                    consumer.accept(future.get());
                } catch (InterruptedException | ExecutionException e) {
                    throw new RuntimeException(e);
                }
            });
        }
    }
}
