package com.chelvc.framework.base.context;

import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.function.Function;
import javax.servlet.ServletRequestEvent;
import javax.servlet.ServletRequestListener;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.model.Platform;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.base.model.Version;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import lombok.NonNull;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

/**
 * 会话上下文工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Component
public class SessionContextHolder implements ServletRequestListener {
    /**
     * 身份标识请求头
     */
    public static final String HEADER_ID = "Id";

    /**
     * 会话类型请求头
     */
    public static final String HEADER_TYPE = "Type";

    /**
     * 租户标识请求头
     */
    public static final String HEADER_TENANT = "Tenant";

    /**
     * 设备标识请求头
     */
    public static final String HEADER_DEVICE = "Device";

    /**
     * 渠道来源请求头
     */
    public static final String HEADER_CHANNEL = "Channel";

    /**
     * 是否首次请求头
     */
    public static final String HEADER_INITIAL = "Initial";

    /**
     * 应用平台请求头
     */
    public static final String HEADER_PLATFORM = "Platform";

    /**
     * 终端编号请求头
     */
    public static final String HEADER_TERMINAL = "Terminal";

    /**
     * 终端版本号请求头
     */
    public static final String HEADER_VERSION = "Version";

    /**
     * 终端序号请求头
     */
    public static final String HEADER_SEQUENCE = "Sequence";

    /**
     * 签名信息请求头
     */
    public static final String HEADER_SIGNATURE = "Signature";

    /**
     * 空会话对象实例
     */
    private static final Session EMPTY_SESSION = new Session();

    /**
     * 对象缓存映射上下文
     */
    private static final ThreadLocal<Map<Object, Object>> CACHE_CONTEXT = ThreadLocal.withInitial(HashMap::new);

    /**
     * 用户会话上下文
     */
    private static final ThreadLocal<Deque<Session>> SESSION_CONTEXT = ThreadLocal.withInitial(ArrayDeque::new);

    /**
     * 获取会话信息
     *
     * @return 会话信息
     */
    public static Session getSession() {
        return getSession(false);
    }

    /**
     * 设置会话信息
     *
     * @param session 会话信息
     */
    public static void setSession(Session session) {
        SESSION_CONTEXT.get().push(ObjectUtils.ifNull(session, EMPTY_SESSION));
    }

    /**
     * 获取会话信息
     *
     * @param nullable 会话是否可以为空
     * @return 会话信息
     */
    public static Session getSession(boolean nullable) {
        Session session = SESSION_CONTEXT.get().peek();
        if (session == EMPTY_SESSION) {
            session = null;
        }
        return nullable ? session : Objects.requireNonNull(session, "Session has not been initialized");
    }

    /**
     * 获取会话属性
     *
     * @param function 会话属性方法
     * @param <T>      会话对象属性类型
     * @return 会话属性值
     */
    public static <T> T getSessionProperty(@NonNull Function<Session, T> function) {
        return getSessionProperty(function, null);
    }

    /**
     * 获取会话属性
     *
     * @param function     会话属性方法
     * @param defaultValue 默认值
     * @param <T>          会话对象属性类型
     * @return 会话属性值
     */
    public static <T> T getSessionProperty(@NonNull Function<Session, T> function, T defaultValue) {
        return ObjectUtils.ifNull(ObjectUtils.ifNull(getSession(true), function), defaultValue);
    }

    /**
     * 获取缓存对象
     *
     * @param key 缓存标识
     * @param <K> 缓存键类型泛型
     * @param <V> 缓存值得类型泛型
     * @return 对象实例
     */
    public static <K, V> V getCache(@NonNull K key) {
        return getCache(key, k -> null);
    }

    /**
     * 获取缓存对象
     *
     * @param key      缓存标识
     * @param provider 对象提供方法
     * @param <K>      缓存键类型泛型
     * @param <V>      缓存值得类型泛型
     * @return 对象实例
     */
    @SuppressWarnings("unchecked")
    public static <K, V> V getCache(@NonNull K key, @NonNull Function<K, V> provider) {
        return (V) CACHE_CONTEXT.get().computeIfAbsent(key, k -> provider.apply((K) k));
    }

    /**
     * 设置缓存对象
     *
     * @param key   缓存标识
     * @param value 缓存对象
     * @param <K>   缓存键类型泛型
     * @param <V>   缓存值得类型泛型
     */
    public static <K, V> void setCache(@NonNull K key, V value) {
        CACHE_CONTEXT.get().put(key, value);
    }

    /**
     * 移除缓存对象
     *
     * @param key 缓存标识
     */
    public static void removeCache(@NonNull String key) {
        CACHE_CONTEXT.get().remove(key);
    }

    /**
     * 移除缓存对象
     *
     * @param keys 缓存标识数组
     */
    public static void removeCaches(@NonNull String... keys) {
        if (keys.length > 0) {
            Map<Object, Object> caches = CACHE_CONTEXT.get();
            for (String key : keys) {
                caches.remove(key);
            }
        }
    }

    /**
     * 移除缓存对象
     *
     * @param keys 缓存标识集合
     */
    public static void removeCaches(@NonNull Collection<String> keys) {
        if (!CollectionUtils.isEmpty(keys)) {
            Map<Object, Object> caches = CACHE_CONTEXT.get();
            keys.forEach(caches::remove);
        }
    }

    /**
     * 判断当前平台是否是指定平台
     *
     * @param platform 指定平台
     * @return true/false
     */
    public static boolean isPlatform(Platform platform) {
        return platform != null && platform == getSessionProperty(Session::getPlatform);
    }

    /**
     * 判断当前终端是否是指定终端
     *
     * @param terminal 指定终端
     * @return true/false
     */
    public static boolean isTerminal(Terminal terminal) {
        return terminal != null && terminal == getSessionProperty(Session::getTerminal);
    }

    /**
     * 判断当前版本是否是指定版本
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @return true/false
     */
    public static boolean isVersion(Terminal terminal, String version) {
        if (terminal == null || StringUtils.isEmpty(version)) {
            return false;
        }
        Session session = getSession(true);
        return terminal == ObjectUtils.ifNull(session, Session::getTerminal)
                && version.equals(ObjectUtils.ifNull(session, Session::getVersion));
    }

    /**
     * 判断当前版本是否和指定版本相同
     *
     * @param version 指定版本注解信息
     * @return true/false
     */
    public static boolean isVersion(com.chelvc.framework.base.annotation.Version version) {
        if (version == null) {
            return false;
        }
        Session session = getSession(true);
        String channel = ObjectUtils.ifNull(session, Session::getChannel);
        Platform platform = ObjectUtils.ifNull(session, Session::getPlatform);
        Terminal terminal = ObjectUtils.ifNull(session, Session::getTerminal);
        if ((ArrayUtils.isEmpty(version.channels()) || ArrayUtils.contains(version.channels(), channel))
                && (ArrayUtils.isEmpty(version.platforms()) || ArrayUtils.contains(version.platforms(), platform))
                && (ArrayUtils.isEmpty(version.terminals()) || ArrayUtils.contains(version.terminals(), terminal))) {
            String current = ObjectUtils.ifNull(session, Session::getVersion);
            if (StringUtils.nonEmpty(version.value())) {
                return Objects.equals(version.value(), current);
            }
            return (StringUtils.isEmpty(version.before())
                    || Version.isBefore(current, version.before(), version.contains()))
                    && (StringUtils.isEmpty(version.after())
                    || Version.isAfter(current, version.after(), version.contains()));
        }
        return false;
    }

    /**
     * 判断当前版本是否在指定版本之前
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @return true/false
     */
    public static boolean isBeforeVersion(Terminal terminal, String version) {
        return isBeforeVersion(terminal, version, false);
    }

    /**
     * 判断当前版本是否在指定版本之前
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @param contains 是否包含当前版本号
     * @return true/false
     */
    public static boolean isBeforeVersion(Terminal terminal, String version, boolean contains) {
        if (terminal == null || StringUtils.isEmpty(version)) {
            return false;
        }
        Session session = getSession(true);
        return terminal == ObjectUtils.ifNull(session, Session::getTerminal)
                && Version.isBefore(ObjectUtils.ifNull(session, Session::getVersion), version, contains);
    }

    /**
     * 判断当前版本是否在指定版本之后
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @return true/false
     */
    public static boolean isAfterVersion(Terminal terminal, String version) {
        return isAfterVersion(terminal, version, false);
    }

    /**
     * 判断当前版本是否在指定版本之后
     *
     * @param terminal 指定终端
     * @param version  指定版本
     * @param contains 是否包含当前版本号
     * @return true/false
     */
    public static boolean isAfterVersion(Terminal terminal, String version, boolean contains) {
        if (terminal == null || StringUtils.isEmpty(version)) {
            return false;
        }
        Session session = getSession(true);
        return terminal == ObjectUtils.ifNull(session, Session::getTerminal)
                && Version.isAfter(ObjectUtils.ifNull(session, Session::getVersion), version, contains);
    }

    /**
     * 判断主体角色是否是管理员
     *
     * @return true/false
     */
    public static boolean isAdmin() {
        return ObjectUtils.ifNull(getSession(true), Session::getType) == Session.Type.ADMIN;
    }

    /**
     * 判断主体是否是普通用户
     *
     * @return true/false
     */
    public static boolean isNormal() {
        return ObjectUtils.ifNull(getSession(true), Session::getType) == Session.Type.NORMAL;
    }

    /**
     * 判断主体是否是客户端
     *
     * @return true/false
     */
    public static boolean isClient() {
        return ObjectUtils.ifNull(getSession(true), Session::getType) == Session.Type.CLIENT;
    }

    /**
     * 判断是否是审核版本
     *
     * @return true/false
     */
    public static boolean isTesting() {
        return isTesting(null, null);
    }

    /**
     * 判断是否是审核版本
     *
     * @param platform 指定平台
     * @return true/false
     */
    public static boolean isTesting(Platform platform) {
        return isTesting(platform, null);
    }

    /**
     * 判断是否是审核版本
     *
     * @param terminal 指定终端
     * @return true/false
     */
    public static boolean isTesting(Terminal terminal) {
        return isTesting(null, terminal);
    }

    /**
     * 判断是否是审核版本
     *
     * @param platform 指定平台
     * @param terminal 指定终端
     * @return true/false
     */
    public static boolean isTesting(Platform platform, Terminal terminal) {
        Session session = getSession(true);
        if (session == null) {
            return false;
        }
        int sequence = ApplicationContextHolder.getProperty("platform.test.sequence", int.class, 0);
        return (platform == null || platform == session.getPlatform())
                && (terminal == null || terminal == session.getTerminal())
                && ObjectUtils.ifNull(session.getSequence(), 0) > sequence;
    }

    /**
     * 获取当前请求对象
     *
     * @return Http请求对象
     */
    public static HttpServletRequest getRequest() {
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        return ObjectUtils.ifNull(attributes, ServletRequestAttributes::getRequest);
    }

    /**
     * 获取当前响应对象
     *
     * @return Http响应对象
     */
    public static HttpServletResponse getResponse() {
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        return ObjectUtils.ifNull(attributes, ServletRequestAttributes::getResponse);
    }

    /**
     * 初始化会话主体
     *
     * @param id   主体标识
     * @param type 会话类型
     * @return 会话信息
     */
    public static Session initializeSessionPrincipal(@NonNull Long id, @NonNull Session.Type type) {
        Session current = Objects.requireNonNull(SESSION_CONTEXT.get().poll(), "Session has not been initialized");
        Session session = Session.builder().id(id).type(type).host(current.getHost()).tenant(current.getTenant())
                .device(current.getDevice()).channel(current.getChannel()).version(current.getVersion())
                .initial(current.getInitial()).platform(current.getPlatform()).terminal(current.getTerminal())
                .signature(current.getSignature()).sequence(current.getSequence()).build();
        SESSION_CONTEXT.get().push(session);
        return session;
    }

    /**
     * 清空上下文信息
     */
    public static void clearSessionContext() {
        // 清空缓存
        CACHE_CONTEXT.remove();

        // 清空会话
        Deque<Session> sessions = SESSION_CONTEXT.get();
        sessions.poll();
        if (sessions.isEmpty()) {
            SESSION_CONTEXT.remove();
        }
    }

    /**
     * 获取客户端身份标识
     *
     * @param request Http请求对象
     * @return 身份标识
     */
    protected Long getId(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_ID), null), Long::parseLong);
    }

    /**
     * 获取会话类型
     *
     * @param request Http请求对象
     * @return 会话类型
     */
    protected Session.Type getType(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_TYPE), null), Session.Type::valueOf);
    }

    /**
     * 获取客户端主机地址
     *
     * @param request Http请求对象
     * @return 主机地址
     */
    protected String getHost(HttpServletRequest request) {
        return StringUtils.ifEmpty(HttpUtils.getRequestAddress(request), null);
    }

    /**
     * 获取客户端租户标识
     *
     * @param request Http请求对象
     * @return 租户标识
     */
    protected Integer getTenant(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_TENANT), null), Integer::parseInt);
    }

    /**
     * 获取客户端设备标识
     *
     * @param request Http请求对象
     * @return 设备标识
     */
    protected String getDevice(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_DEVICE), null), String::trim);
    }

    /**
     * 获取客户端渠道来源
     *
     * @param request Http请求对象
     * @return 渠道来源
     */
    protected String getChannel(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_CHANNEL), null), String::trim);
    }

    /**
     * 获取客户端版本号
     *
     * @param request Http请求对象
     * @return 版本号
     */
    protected String getVersion(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_VERSION), null), String::trim);
    }

    /**
     * 获取客户端平台标识
     *
     * @param request Http请求对象
     * @return 平台标识
     */
    protected Platform getPlatform(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_PLATFORM), null), Platform::valueOf);
    }

    /**
     * 获取客户端终端
     *
     * @param request Http请求对象
     * @return 终端标识
     */
    protected Terminal getTerminal(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_TERMINAL), null), Terminal::valueOf);
    }

    /**
     * 获取签名信息
     *
     * @param request Http请求对象
     * @return 签名信息
     */
    protected String getSignature(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_SIGNATURE), null), String::trim);
    }

    /**
     * 获取是否首次请求
     *
     * @param request Http请求对象
     * @return true/false
     */
    protected Boolean getInitial(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_INITIAL), null), Boolean::parseBoolean);
    }

    /**
     * 获取客户端序号
     *
     * @param request Http请求对象
     * @return 客户端序号
     */
    protected Integer getSequence(HttpServletRequest request) {
        return ObjectUtils.ifNull(StringUtils.ifEmpty(request.getHeader(HEADER_SEQUENCE), null), Integer::parseInt);
    }

    @Override
    public void requestInitialized(ServletRequestEvent event) {
        HttpServletRequest request = (HttpServletRequest) event.getServletRequest();
        Session session = Session.builder().id(this.getId(request)).type(this.getType(request))
                .host(this.getHost(request)).tenant(this.getTenant(request)).device(this.getDevice(request))
                .channel(this.getChannel(request)).version(this.getVersion(request)).initial(this.getInitial(request))
                .platform(this.getPlatform(request)).terminal(this.getTerminal(request))
                .signature(this.getSignature(request)).sequence(this.getSequence(request)).build();
        setSession(session);
    }

    @Override
    public void requestDestroyed(ServletRequestEvent event) {
        clearSessionContext();
    }
}
