package com.chelvc.framework.security.token;

import java.util.Objects;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.security.config.WebSecurityConfigurer;
import com.chelvc.framework.security.util.TokenUtils;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.http.HttpStatus;
import org.springframework.security.oauth2.core.OAuth2AuthenticationException;
import org.springframework.security.oauth2.core.OAuth2Error;
import org.springframework.security.oauth2.core.OAuth2TokenValidatorResult;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.security.oauth2.server.resource.BearerTokenError;
import org.springframework.util.StringUtils;

/**
 * 基于Redis的令牌有效性验证实现
 *
 * @author Woody
 * @date 2023/4/5
 * @see TokenActiveValidator
 * @see WebSecurityConfigurer
 */
@Slf4j
public class RedisTokenActiveValidator implements TokenActiveValidator {
    private final RedisTemplate<String, Object> redisTemplate;

    public RedisTokenActiveValidator(@NonNull RedisTemplate<String, Object> redisTemplate) {
        this.redisTemplate = redisTemplate;
    }

    @Override
    public OAuth2TokenValidatorResult validate(@NonNull Jwt jwt) {
        // 验证Redis中是否存在该令牌，同时忽略Redis请求/操作异常（为了提高系统可用性）
        String token = null;
        Terminal terminal = SessionContextHolder.getSessionProperty(Session::getTerminal);
        String key = TokenUtils.key(terminal, TokenUtils.getId(jwt));
        if (log.isDebugEnabled()) {
            log.debug("Validating redis token key: {}", key);
        }
        try {
            if (StringUtils.isEmpty(token = (String) this.redisTemplate.opsForValue().get(key))) {
                return OAuth2TokenValidatorResult.failure(
                        new OAuth2Error("invalid_token", "Token is not active", null)
                );
            }
        } catch (Exception e) {
            log.warn("Token validate failed: {}", e.getMessage());
        }

        // 判断令牌是否相同，如果不相同则标识令牌已经被重置（账号已在其他地方登录）
        if (!StringUtils.isEmpty(token) && !Objects.equals(token, jwt.getTokenValue())) {
            throw new OAuth2AuthenticationException(
                    new BearerTokenError("invalid_token", HttpStatus.GONE, "Token has been reset", null)
            );
        }
        return OAuth2TokenValidatorResult.success();
    }
}
