package com.chelvc.framework.security.util;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.google.common.collect.Sets;
import org.springframework.security.oauth2.common.OAuth2AccessToken;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.util.CollectionUtils;

/**
 * 令牌工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
public final class TokenUtils {
    /**
     * 主体身份标识
     */
    public static final String JWT_CLAIM_ID = "id";

    /**
     * JWT身份标识
     */
    public static final String JWT_CLAIM_JTI = "jti";

    /**
     * 会话类型标识
     */
    public static final String JWT_CLAIM_TYPE = "type";

    /**
     * JWT手机号标识
     */
    public static final String JWT_CLAIM_MOBILE = "mobile";

    /**
     * 用户账号标识
     */
    public static final String JWT_CLAIM_USER_NAME = "user_name";

    /**
     * 客户端账号标识
     */
    public static final String JWT_CLAIM_CLIENT_ID = "client_id";

    /**
     * 用户权限标识
     */
    public static final String JWT_CLAIM_AUTHORITIES = "authorities";

    private TokenUtils() {
    }

    /**
     * 获取主体身份
     *
     * @param jwt JWT对象
     * @return 主体身份标识
     */
    public static Long getId(Jwt jwt) {
        return ObjectUtils.ifNull(jwt, t -> ObjectUtils.object2long(t.getClaim(JWT_CLAIM_ID)));
    }

    /**
     * 获取主体身份
     *
     * @param token OAuth2访问令牌对象
     * @return 主体身份标识
     */
    public static Long getId(OAuth2AccessToken token) {
        return ObjectUtils.ifNull(
                ObjectUtils.ifNull(token, OAuth2AccessToken::getAdditionalInformation),
                info -> ObjectUtils.object2long(info.get(JWT_CLAIM_ID))
        );
    }

    /**
     * 获取JWT身份标识
     *
     * @param jwt JWT对象
     * @return JWT身份标识
     */
    public static String getJti(Jwt jwt) {
        return jwt == null ? null : jwt.getClaimAsString(JWT_CLAIM_JTI);
    }

    /**
     * 获取会话类型
     *
     * @param jwt JWT对象
     * @return 会话类型
     */
    public static Session.Type getType(Jwt jwt) {
        String type = jwt == null ? null : jwt.getClaimAsString(JWT_CLAIM_TYPE);
        return StringUtils.isEmpty(type) ? null : Session.Type.valueOf(type);
    }

    /**
     * 获取用户标识
     *
     * @param jwt JWT对象
     * @return 用户标识
     */
    public static String getUser(Jwt jwt) {
        return jwt == null ? null : jwt.getClaimAsString(JWT_CLAIM_USER_NAME);
    }

    /**
     * 获取手机号
     *
     * @param jwt JWT对象
     * @return 手机号
     */
    public static String getMobile(Jwt jwt) {
        return jwt == null ? null : jwt.getClaimAsString(JWT_CLAIM_MOBILE);
    }

    /**
     * 获取客户端标识
     *
     * @param jwt JWT对象
     * @return 客户端标识
     */
    public static String getClient(Jwt jwt) {
        return jwt == null ? null : jwt.getClaimAsString(JWT_CLAIM_CLIENT_ID);
    }

    /**
     * 获取用户权限集合
     *
     * @param jwt JWT对象
     * @return 权限标识集合
     */
    public static Set<String> getAuthorities(Jwt jwt) {
        List<String> authorities = jwt == null ? null : jwt.getClaimAsStringList(JWT_CLAIM_AUTHORITIES);
        return CollectionUtils.isEmpty(authorities) ? Collections.emptySet() : Sets.newHashSet(authorities);
    }

    /**
     * 构建令牌缓存标识（访问端+主体标识）
     *
     * @param terminal 客户终端
     * @param id       主体标识
     * @return 标识
     */
    public static String key(Terminal terminal, Long id) {
        return String.format("token:%s:%d", terminal, id);
    }
}
